#!/usr/bin/env python3

"""
pydepict.utils

Utility functions

Copyright (c) 2022 William Lee and The University of Sheffield. See LICENSE for details
"""

import datetime as dt
from itertools import product
from typing import Dict, Iterable, List, Optional, Tuple, TypeVar, Union

import networkx as nx

from .consts import CHAIN_ELEMENTS, AtomAttribute, BondAttribute, GraphCoordinates
from .models import Vector

__all__ = [
    "bond_order_sum",
    "atom_valence",
    "is_allenal_center",
    "depicted_distance",
    "average_depicted_bond_length",
    "none_iter",
]

T = TypeVar("T")


def get_atom_attrs(
    atom_index: int, graph: nx.Graph, *attrs: str, allow_none: bool = False
) -> Union[AtomAttribute, Tuple[AtomAttribute, ...]]:
    """
    Retrieves atom attributes for the specified atom in the specified graph.

    Accepts one or more attribute names. If one attribute name is provided,
    then that attribute value is returned. If multiple attributes name are provided,
    then a tuple of attribute values is returned, in the order in which
    the attribute names are passed.

    :param atom_index: The index of the atom to look for
    :type atom_index: int
    :param graph: The graph to look for the atom in
    :type graph: nx.Graph
    :param *args: One or more attribute names to fetch
    :type *args: str
    :param allow_none: Whether :data:`None` may be returned
                       if an attribute is not found, defaults to :data:`False`
    :type allow_none: bool
    :return: The attribute value, or a tuple of attribute values
    :rtype: Union[AtomAttribute, Tuple[AtomAttribute, ...]]
    """
    atom = graph.nodes[atom_index]
    if len(attrs) == 1:
        return atom.get(attrs[0], None) if allow_none else atom[attrs[0]]
    return (
        tuple(atom.get(attr, None) for attr in attrs)
        if allow_none
        else tuple(atom[attr] for attr in attrs)
    )


def get_bond_attrs(
    u: int, v: int, graph: nx.Graph, *attrs: str, allow_none: bool = True
) -> Union[BondAttribute, Tuple[BondAttribute, ...]]:
    """
    Retrieves bond attributes between the specified atoms in the specified graph.

    Accepts one or more attribute names. If one attribute name is provided,
    then that attribute value is returned. If multiple attributes name are provided,
    then a tuple of attribute values is returned, in the order in which
    the attribute names are passed.

    :param u: One endpoint of the bond
    :type u: int
    :param v: The other endpoint of the bond
    :type v: int
    :param graph: The graph to look for the atom in
    :type graph: nx.Graph
    :param *args: One or more attribute names to fetch
    :type *args: str
    :param allow_none: Whether :data:`None` may be returned
                       if an attribute is not found, defaults to :data:`False`
    :type allow_none: bool
    :return: The attribute value, or a tuple of attribute values
    :rtype: Union[AtomAttribute, Tuple[AtomAttribute, ...]]
    """
    bond = graph[u][v]
    if len(attrs) == 1:
        return bond.get(attrs[0], None) if allow_none else bond[attrs[0]]
    return (
        tuple(bond.get(attr, None) for attr in attrs)
        if allow_none
        else tuple(bond[attr] for attr in attrs)
    )


def bond_order_sum(atom_index: int, graph: nx.Graph) -> int:
    """
    Calculates the sum of the orders of the bonds adjacent to the atom
    at the specified index within the specified graph.

    :param atom_index: The index of the atom to find the bond order sum for
    :type atom_index: int
    :param graph: The graph to look for the atom in
    :type graph: nx.Graph
    :return: The sum of the orders of the bonds adjacent to the specified atom
    :type: int
    """
    return sum(bond["order"] for bond in graph.adj[atom_index].values())


def atom_valence(atom_index: int, graph: nx.Graph) -> int:
    """
    Calculates the valence of the atom with the specified index
    within the specified graph.

    :param atom_index: The index of the atom to calculate valence for
    :type atom_index: int
    :param graph: The graph to look for the atom in
    :type graph: nx.Graph
    :return: The valence of the atom
    :type: int
    """
    atom = graph.nodes[atom_index]
    hcount = getattr(atom, "hcount", None)

    return (
        bond_order_sum(atom_index, graph)
        + (0 if hcount is None else hcount)
        + atom["charge"]
    )


def neighbors(
    atom_index: int, graph: nx.Graph, exclude: Iterable[int] = []
) -> List[int]:
    """
    Determines the neighbors of the specified atom,
    returned as a list of indices.

    :param atom_index: The index of the atom to find neighbors for
    :type atom_index: int
    :param graph: The graph to look for the atom in
    :type graph: nx.Graph
    :param exclude: A list of indices to exclude from neighbors
    :type exclude: Iterable[int]
    :return: A list of neighbors
    :rtype: List[int]
    """
    return list(filter(lambda v: v not in exclude, graph[atom_index].keys()))


def num_heavy_atom_neighbors(atom_index: int, graph: nx.Graph) -> int:
    """
    Determines the number of heavy atom neighbors for the specified atom.
    A heavy atom is any atom that is not hydrogen.

    :param atom_index: The index of the atom to find
                       the number of heavy atom neighbors for
    :type atom_index: int
    :param graph: The graph to look for the atom in
    :type graph: nx.Graph
    :return: The number of heavy atom neighbors
    :rtype: int
    """
    return sum(
        get_atom_attrs(v, graph, "element") != "H" for v in neighbors(atom_index, graph)
    )


def num_bond_order(atom_index: int, graph: nx.Graph, order: int) -> bool:
    """
    Counts the number of bonds of a particular order.

    :param atom_index: The index of the atom to count for
    :type atom_index: int
    :param graph: The graph to look for the atom in
    :type graph: nx.Graph
    :param order: The order of bonds to count
    :type order: int
    :return: The number of bonds of the specified order
    :rtype: int
    """
    return list(bond["order"] == 3 for bond in graph[atom_index].values()).count(order)


def is_allenal_center(atom_index: int, graph: nx.Graph) -> bool:
    """
    Determines whether or not the atom at the specified index
    within the specified graph is the carbon at the center of an allene.

    :param atom_index: The index of the atom to calculate valence for
    :type atom_index: int
    :param graph: The graph to look for the atom in
    :type graph: nx.Graph
    :return: Whether the atom is an allene center or not
    :type: bool
    """
    return (
        get_atom_attrs(atom_index, graph, "element") == "C"
        and num_bond_order(atom_index, graph, 2) == 2
        and bond_order_sum(atom_index, graph) == 4
    )


def is_chain_atom(atom_index: int, graph: nx.Graph) -> bool:
    """
    Determines whether or not the specified atom in the specified graph
    is a chain atom (i.e. eligible to be treated as being in a chain).

    :param graph: The graph to look for the atom in
    :type graph: nx.Graph
    :param atom_index: The index of the atom
    :type atom_index: int
    :return: Whether the atom is a chain atom
    :rtype: bool
    """
    element, charge = get_atom_attrs(atom_index, graph, "element", "charge")
    return (
        element in CHAIN_ELEMENTS
        and charge == 0
        and num_bond_order(atom_index, graph, 3) == 0
        and num_heavy_atom_neighbors(atom_index, graph) in (2, 3)
    )


def depicted_distance(u: int, v: int, positions: GraphCoordinates) -> float:
    """
    Calculates the depicted distance between two atoms in a molecule graph.

    :param u: The index of one atom
    :type u: int
    :param v: The index of the other atom
    :type v: int
    :param positions: The dictionary of depiction coordinates
    :type positions: GraphCoordinates
    :return: The distance between the two atoms
    :rtype: float
    """
    coords1, coords2 = positions[u], positions[v]

    return Vector.distance(coords1, coords2)


def average_depicted_bond_length(graph: nx.Graph, positions: GraphCoordinates) -> float:
    """
    Calculates the average length of the representation of bonds
    in a graph that has depiction coordinates.

    Average length is calculated by calculating a mean of the Euclidean distance
    between the two endpoints of each edge.

    :param graph: The graph to calculate the average bond length for
    :type graph: nx.Graph
    :param positions: The dictionary of depiction coordinates
    :type positions: GraphCoordinates
    :return: The average bond length in the graph, which is 0 if the graph has no edges
    :rtype: float
    """
    if not graph.edges:
        return 0
    total_distance = sum(depicted_distance(u, v, positions) for u, v in graph.edges)
    return total_distance / len(graph.edges)


def depiction_width(sample: Dict[int, Vector]) -> float:
    """
    Calculates the depicted width of a depiction sample,
    calculated as the difference between the lowest and highest x coordinates.

    :param sample: The depiction sample
    :type sample: Dict[int, Vector]
    :return: The depicted width
    :rtype: float
    """
    min_x = min(vector.x for vector in sample.values())
    max_x = max(vector.x for vector in sample.values())
    return max_x - min_x


def none_iter(iterable: Iterable[T]) -> Iterable[Iterable[Optional[T]]]:
    """
    Takes an iterable, and returns an iterable that iterates all possible
    substitutions of the elements of the iterable with :data:`None`.

    :param iterable: The iterable
    :type iterable: Iterable[T]
    :return: The iterable of iterables with all possible substitutions
    :rtype: Iterable[Iterable[Optional[T]]]
    """
    return product(
        *((value, None) if value is not None else (None,) for value in iterable)
    )


def prune_hydrogens(graph: nx.Graph, atoms: List[int]):
    """
    Finds hydrogen atoms in the graph, and removes them from the specified list
    of atom indices.

    If removing an index would empty the list, then one remaining index is kept.

    :param graph: The graph to find hydrogens within
    :type graph: nx.Graph
    :param atoms: The list to remove atom indices from.
    :type atoms: List[int]
    """
    for atom_index, element in dict(graph.nodes(data="element")).items():
        if element == "H" and atom_index in atoms and len(atoms) > 1:
            atoms.remove(atom_index)


def prune_terminals(graph: nx.Graph, atoms: List[int]):
    """
    Finds terminals atoms in the graph, and removes them from the specified list
    of atom indices.

    If removing an index would empty the list, then one remaining index is kept.

    :param graph: The graph to find terminals within
    :type graph: nx.Graph
    :param atoms: The list to remove atom indices from.
    :type atoms: List[int]
    """
    for atom_index in list(graph.nodes):
        if len(graph[atom_index]) <= 1 and atom_index in atoms and len(atoms) > 1:
            atoms.remove(atom_index)


def get_datetime_filename() -> str:
    """
    Returns a human-readable, datetime-based filename without the extension.

    :return: The filename
    :rtype: str
    """
    return dt.datetime.now().strftime("pydepict_%Y_%m_%d_%H_%M_%S")



#!/usr/bin/env python3

"""
pydepict.consts

Constants for parsing, depicting and rendering

Copyright (c) 2022 William Lee and The University of Sheffield. See LICENSE for details
"""

from math import pi
from typing import Dict, List, Optional, Tuple, Union

from .models import Vector

# CUSTOM TYPES

ChiralSpec = Optional[Tuple[Optional[str], int]]
BondAttribute = Optional[Union[bool, str, int, float]]
AtomAttribute = Optional[Union[bool, str, int, float, ChiralSpec]]
Atom = Dict[str, AtomAttribute]
Bond = Dict[str, BondAttribute]
Chain = Tuple[List[Atom], List[Bond]]
Rnum = Tuple[int, Optional[float]]
Rnums = Dict[int, Rnum]
AtomRnums = List[Tuple[int, Optional[float]]]
NeighborSpec = Tuple[Optional[str], Optional[float]]
AtomPattern = Dict[NeighborSpec, Tuple[Vector, ...]]
AtomPatterns = List[Tuple[AtomPattern, float]]
NeighborConstraints = Dict[int, Vector]
ChainPattern = Tuple[Tuple[Vector, Vector], Dict[int, Tuple[Vector, ...]]]
ConstraintsCandidates = Dict[
    Tuple[int],
    Tuple[
        List[List[NeighborConstraints]],
        List[float],
    ],
]
GraphCoordinates = Dict[int, Vector]

# GENERAL CHEMISTRY DATA

HALOGENS = frozenset("F Cl Br I At".split())

# GEOMETRY

THIRTY_DEGREES = pi / 6

# PARSER ELEMENT SYMBOLS

WILDCARD = "*"

STANDARD_SYMBOLS = frozenset(
    (
        "H He Li Be B C N O F Ne Na Mg Al Si P S Cl Ar K Ca Sc Ti V Cr Mn Fe Co Ni Cu "
        "Zn Ga Ge As Se Br Kr Rb Sr Y Zr Nb Mo Tc Ru Rh Pd Ag Cd In Sn Sb Te I Xe Cs "
        "Ba La Ce Pr Nd Pm Sm Eu Gd Tb Dy Ho Er Tm Yb Lu Hf Ta W Re Os Ir Pt Au Hg Tl "
        "Pb Bi Po At Rn Fr Ra Ac Th Pa U Np Pu Am Cm Bk Cf Es Fm Md No Lr Rf Db Sg Bh "
        "Hs Mt Ds Rg Cn Nh Fl Mc Lv Ts Og Br"
    ).split()
)
AROMATIC_SYMBOLS = frozenset("b c n o s p se as".split())
ELEMENT_SYMBOLS = STANDARD_SYMBOLS | AROMATIC_SYMBOLS | {WILDCARD}

STANDARD_ORGANIC_SYMBOLS = frozenset("B C N O S P F Cl Br I".split())
AROMATIC_ORGANIC_SYMBOLS = frozenset("b c n o s p".split())
ORGANIC_SYMBOLS = STANDARD_ORGANIC_SYMBOLS | AROMATIC_ORGANIC_SYMBOLS | {WILDCARD}

ELEMENT_SYMBOL_FIRST_CHARS = frozenset(element[0] for element in ELEMENT_SYMBOLS)
ORGANIC_SYMBOL_FIRST_CHARS = frozenset(element[0] for element in ORGANIC_SYMBOLS)

# PARSER CHIRALITY

CHIRALITY_CODES = frozenset("TH AL SP TB OH".split())
CHIRALITY_CODES_FIRST_CHARS = frozenset(code[0] for code in CHIRALITY_CODES)
CHIRALITY_RANGES = {
    "TH": 2,
    "AL": 2,
    "SP": 3,
    "TB": 20,
    "OH": 30,
}

# OTHER PARSER SYMBOLS

CHARGE_SYMBOLS = frozenset({"-", "+"})
TERMINATORS = frozenset({" ", "\t", "\r", "\n"})
BOND_TO_ORDER: Dict[str, float] = {
    "-": 1,
    "=": 2,
    "#": 3,
    "$": 4,
    "/": 1,
    "\\": 1,
}

# PARSER SPECIFICATIONS

MIN_CHARGE_MAGNITUDE = 15
VALENCES: Dict[str, Optional[Tuple[int, ...]]] = {
    "B": (3,),
    "C": (4,),
    "N": (3, 5),
    "O": (2,),
    "P": (3, 5),
    "S": (2, 4, 6),
    "F": (1,),
    "Cl": (1,),
    "Br": (1,),
    "*": None,
}

# PARSER TEMPLATES

DEFAULT_ATOM = {
    "isotope": None,
    "element": "*",
    "hcount": 0,
    "charge": 0,
    "class": None,
    "aromatic": False,
}
DEFAULT_BOND = {"order": 1}

# DEPICTER VECTORS

RRR = Vector(1, 0)
RRU = Vector(1, 0).rotate(THIRTY_DEGREES)
RUU = Vector(1, 0).rotate(THIRTY_DEGREES * 2)
UUU = Vector(1, 0).rotate(THIRTY_DEGREES * 3)
LUU = Vector(1, 0).rotate(THIRTY_DEGREES * 4)
LLU = Vector(1, 0).rotate(THIRTY_DEGREES * 5)
LLL = Vector(1, 0).rotate(THIRTY_DEGREES * 6)
LLD = Vector(1, 0).rotate(THIRTY_DEGREES * 7)
LDD = Vector(1, 0).rotate(THIRTY_DEGREES * 8)
DDD = Vector(1, 0).rotate(THIRTY_DEGREES * 9)
RDD = Vector(1, 0).rotate(THIRTY_DEGREES * 10)
RRD = Vector(1, 0).rotate(THIRTY_DEGREES * 11)

# DEPICTER CONSTRAINTS

ATOM_PATTERNS: Dict[Optional[str], AtomPatterns] = {
    "C": [
        (
            {
                (None, None): (LLL,),
            },
            1,
        ),
        (
            {
                (None, 1): (LLD,),
                (None, 2): (RRD,),
            },
            1,
        ),
        (
            {
                ("C", 1): (LLD,),
                (None, None): (RRD,),
            },
            1,
        ),
        (
            {
                (None, 1): (LLL,),
                (None, 3): (RRR,),
            },
            1,
        ),
        (
            {
                ("C", 2): (LLL, RRR),
            },
            1,
        ),
        (
            {
                (None, 1): (LLD, RRD),
            },
            1,
        ),
        (
            {
                (None, 1): (LUU, LDD),
                ("C", 2): (RRR,),
            },
            1,
        ),
        (
            {
                (None, 1): (LLD, UUU, RRD),
            },
            1,
        ),
        (
            {
                ("O", None): (UUU,),
                (None, 1): (LLD, RRD),
            },
            1,
        ),
        (
            {
                ("O", None): (LLD,),
                (None, 1): (UUU, RRD),
            },
            1,
        ),
        (
            {
                ("C", 1): (RRR, DDD, LLL, UUU),
            },
            1,
        ),
        (
            {
                ("C", 1): (RRR, RDD, LLL, RUU),
            },
            1,
        ),
        (
            {
                ("C", 1): (LLD,),
                (None, 1): (RRD,),
                ("X", 1): (LUU, RUU),
            },
            1,
        ),
        (
            {
                ("X", 1): (RRR, RDD, RUU),
                ("C", 1): (LLL,),
            },
            1,
        ),
        (
            {
                ("C", 1): (LLD, RRD),
                (None, 1): (LUU, RUU),
            },
            1,
        ),
        (
            {
                (None, 1): (UUU, RRR),
                ("C", 1): (DDD, LLL),
            },
            0.2,
        ),
        (
            {
                (None, 1): (UUU, DDD),
                ("C", 1): (LLL, RRR),
            },
            0.2,
        ),
    ],
    "N": [
        (
            {
                (None, 1): (LLD, RRD),
            },
            1,
        ),
        (
            {
                (None, 2): (LLD,),
                (None, 1): (RRD,),
            },
            1,
        ),
        (
            {
                (None, 2): (LLL,),
                (None, 1): (RRR,),
            },
            0.1,
        ),
    ],
    "O": [
        (
            {
                (None, 1): (LLD, RRD),
            },
            1,
        ),
        (
            {
                (None, 1): (LLL, RRR),
            },
            0.1,
        ),
    ],
    "P": [
        (
            {
                (None, 1): (LLD, RRD),
            },
            1,
        ),
        (
            {
                (None, 1): (LLL, RRR),
            },
            0.2,
        ),
        (
            {
                (None, 1): (LLL, RRR),
                ("O", 2): (UUU, DDD),
            },
            1,
        ),
        (
            {
                (None, 1): (RRR, DDD),
                ("O", 2): (LLL, UUU),
            },
            0.1,
        ),
    ],
    "S": [
        (
            {
                (None, 1): (LLD, RRD),
            },
            1,
        ),
        (
            {
                (None, 1): (LLL, RRR),
            },
            0.4,
        ),
        (
            {
                (None, 1): (LLL, RRR),
                ("O", 2): (UUU, DDD),
            },
            1,
        ),
        (
            {
                (None, 1): (RRR, DDD),
                ("O", 2): (LLL, UUU),
            },
            0.1,
        ),
    ],
    "Se": [
        (
            {
                (None, 1): (LLD, RRD),
            },
            1,
        ),
        (
            {
                (None, 1): (LLL, RRR),
            },
            0.5,
        ),
    ],
    "Si": [
        (
            {
                (None, 1): (LLD, RRD),
            },
            1,
        ),
        (
            {
                (None, 1): (LLL, RRR),
            },
            0.5,
        ),
    ],
    None: [
        (
            {
                (None, None): (LLL,),
            },
            1,
        ),
        (
            {
                (None, None): (LLL, RRR),
            },
            1,
        ),
        (
            {
                (None, 1): (LLD, RRD),
            },
            1,
        ),
        (
            {
                (None, None): (LLL, RUU, RDD),
            },
            1,
        ),
        (
            {
                (None, None): (UUU, RRR, DDD, LLL),
            },
            1,
        ),
    ],
}

# Calculates reflected atom constraints
for meth in (Vector.x_reflect, Vector.y_reflect):
    for patterns in ATOM_PATTERNS.values():
        patterns_copy = patterns.copy()
        patterns.extend(
            (
                {
                    atom: tuple(meth(v) for v in vectors)
                    for atom, vectors in pattern.items()
                },
                weight,
            )
            for pattern, weight in patterns_copy
        )
del meth, patterns, patterns_copy

CHAIN_PATTERN_UNITS: List[Tuple[ChainPattern, ChainPattern]] = [
    (
        (
            (LLD, RRD),
            {
                1: (UUU,),
                2: (LUU, RUU),
            },
        ),
        (
            (LLU, RRU),
            {
                1: (DDD,),
                2: (LDD, RDD),
            },
        ),
    ),
    (
        (
            (DDD, RRU),
            {
                1: (LLU,),
                2: (LLL, LUU),
            },
        ),
        (
            (LLD, UUU),
            {
                1: (RRD,),
                2: (RRR, RDD),
            },
        ),
    ),
]

# Combine reflections and order swapping to produce other patterns

for pattern in CHAIN_PATTERN_UNITS.copy():
    first, second = pattern
    swapped = (second, first)
    if swapped not in CHAIN_PATTERN_UNITS:
        CHAIN_PATTERN_UNITS.append(swapped)
for meth in (Vector.x_reflect, Vector.y_reflect):
    for pattern in CHAIN_PATTERN_UNITS.copy():
        new_pattern = tuple(
            (
                (meth(prev_vector), meth(next_vector)),
                {
                    num_subs: tuple(meth(v) for v in vectors)
                    for num_subs, vectors in sub_constraints.items()
                },
            )
            for (prev_vector, next_vector), sub_constraints in pattern
        )
        if new_pattern not in CHAIN_PATTERN_UNITS:
            CHAIN_PATTERN_UNITS.append(new_pattern)

# OTHER DEPICTER CONSTANTS

CHAIN_ELEMENTS = frozenset("C N O S".split())
SAMPLE_SIZE = 100
EPSILON = 0.0001

# RENDERER WINDOW ATTRIBUTES

WINDOW_TITLE = "pydepict"

# RENDERER COLORS

WHITE = (255, 255, 255)
BLACK = (0, 0, 0)

# RENDERER DIMENSIONS

MIN_DISPLAY_SIZE = (400, 300)
FRAME_MARGIN = 50
DISPLAY_BOND_LENGTH = 55
BOND_WIDTH = 6
TEXT_MARGIN = 2

# RENDERER FONT

FONT_FAMILY = "Arial"
FONT_SIZE = 20

# RENDERER PATHS
SCREENSHOTS_DIR = "screenshots"



#!/usr/bin/env python3

"""
pydepict.__main__

A program for parsing and rendering chemical structures from SMILES strings.

Copyright (c) 2022 William Lee and The University of Sheffield. See LICENSE for details
"""


import argparse
import traceback
from tkinter import Event, Tk
from tkinter.ttk import Button, Entry, Frame, Label

from . import Renderer, depict, parse, show

PADDING = 10

__all__ = ["Program", "main"]


class Program:
    """
    Class representing a program for accepting SMILES input,
    and displaying the corresponding diagram.

    The program is built using :module:`tkinter`.
    """

    def __init__(self):
        self.root = Tk()
        self.root.wm_title("pydepict")

        # Instantiate widgets
        self.frame = Frame(self.root, padding=PADDING)
        self.smiles_input_label = Label(self.frame, text="SMILES")
        self.smiles_input = Entry(self.frame)
        self.smiles_input.bind("<Return>", lambda _: self._show_smiles())
        self.smiles_input.bind("<Control-KeyRelease-a>", self._select_all)
        self.display_button = Button(
            self.frame, text="Display", command=self._show_smiles
        )
        self.error_message = Label(self.frame, foreground="red")

        # Arrange and display widgets
        self.frame.grid()
        self.smiles_input_label.grid(column=0, row=0, padx=PADDING, pady=PADDING)
        self.smiles_input.grid(column=1, row=0, padx=PADDING, pady=PADDING)
        self.display_button.grid(column=0, row=1, columnspan=2, pady=PADDING)
        self.error_message.grid(column=0, row=2, columnspan=2, pady=PADDING)

        # Instantiate renderer
        self.renderer = Renderer()
        self.renderer.show(False)
        # Bind renderer close event to close the program
        self.renderer.on("close", self.root.destroy)

        # Bind window close event
        self.root.protocol("WM_DELETE_WINDOW", self.close)

    @staticmethod
    def _select_all(event: Event):
        event.widget.select_range(0, "end")
        event.widget.icursor("end")

    def _show_smiles(self):
        self.error_message.config(text="")
        smiles = self.smiles_input.get()
        try:
            graph, _ = parse(smiles)
            positions = depict(graph)
            self.renderer.set_structure(graph, positions)
            self.renderer.title = smiles
        except Exception as e:
            self.error_message.config(text=f"{e.__class__.__name__}: {str(e)}")
            traceback.print_exc()

    def close(self):
        """
        Close the program.
        """
        # Remove callback to prevent deadlock
        self.renderer.not_on("close", self.root.destroy)
        self.renderer.close()
        self.root.destroy()

    def run(self):
        """
        Run the program.
        """
        self.root.mainloop()


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument("smiles", nargs="?", default=None)

    return parser.parse_args()


def main():
    """
    Runs the standalone program.

    If the calling script is run with no arguments, then a dialog is shown,
    allowing entry of SMILES strings multiple times.

    If a SMILES string is passed as the first and only argument,
    then only the renderer window for that SMILES string is shown.
    """
    args = parse_args()
    if args.smiles is None:
        Program().run()
    else:
        show(args.smiles)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""
pydepict.parser

Parsing for strings conforming to the OpenSMILES specification

Copyright (c) 2022 William Lee and The University of Sheffield. See LICENSE for details
"""

import string
import warnings
from functools import wraps
from typing import Iterable, List, Optional, Tuple, Type, TypeVar, Union

import networkx as nx

from .consts import (
    BOND_TO_ORDER,
    CHARGE_SYMBOLS,
    CHIRALITY_CODES,
    CHIRALITY_CODES_FIRST_CHARS,
    CHIRALITY_RANGES,
    DEFAULT_ATOM,
    DEFAULT_BOND,
    ELEMENT_SYMBOL_FIRST_CHARS,
    ELEMENT_SYMBOLS,
    ORGANIC_SYMBOL_FIRST_CHARS,
    ORGANIC_SYMBOLS,
    TERMINATORS,
    VALENCES,
    Atom,
    AtomAttribute,
    AtomRnums,
    Bond,
    BondAttribute,
    Rnum,
    Rnums,
)
from .errors import ParserError, ParserWarning
from .models import Stream
from .utils import atom_valence, is_allenal_center

__all__ = ["parse"]

T = TypeVar("T")
E = TypeVar("E", Type[ParserError], Type[ParserWarning])

# Sentinel object
EXPECT_DEFAULT = object()


def fill_hydrogens(graph: nx.Graph):
    """
    Fills the hcount attribute for atoms where it is :data:`None`
    (implies the atom is organic subset).

    :param graph: The graph to fill hydrogens for.
    :type graph: nx.Graph
    """
    for atom_index, attrs in graph.nodes(data=True):
        if attrs["hcount"] is None:
            element = attrs["element"]
            # Get all "normal" valences for the current atom
            element_valences = VALENCES[element]
            if element_valences is None:
                continue
            current_valence = atom_valence(atom_index, graph)
            # Possible valences must be at least the current valence of the atom
            possible_valences = list(
                filter(lambda x: x >= current_valence, element_valences)
            )
            if not possible_valences:
                # Hydrogen count is 0 if current valence
                # is already higher than any known valence
                attrs["hcount"] = 0
            target_valence = min(possible_valences)
            attrs["hcount"] = target_valence - current_valence


def imply_shorthand_chirality(graph: nx.Graph):
    """
    Implies the chirality type for atoms in the specified graph
    where there is shorthand chirality specification, i.e. @ or @@

    Shorthand chirality is implied whenever the chirality code is :data:`None`

    :param graph: The graph to scan for shorthand chirality
    :type graph: nx.Graph
    """

    for atom_index, chirality in graph.nodes(data="chirality"):
        if chirality is not None and chirality[0] is None:
            graph.nodes[atom_index]["chirality"] = (
                "AL" if is_allenal_center(atom_index, graph) else "TH"
            )


def infer_bond(atom_aromatic: bool, other_atom_aromatic: bool) -> Bond:
    """
    Returns the inferred bond in the absence of any bond symbol
    based on the relevant atoms.

    Aromatic bond is inferred between two aromatic atoms,
    otherwise single bond is inferred.

    :param atom_aromatic: The aromaticity of one of the atoms
    :type atom_aromatic: bool
    :param other_atom_aromatic: The aromaticity of the other atom
    :type other_atom_aromatic: bool
    :return: A new bond attribute dictionary
    :rtype: Bond
    """
    return new_bond(order=1.5 if atom_aromatic and other_atom_aromatic else 1)


def resolve_rnums(
    atom_rnums: AtomRnums,
    atom_index: int,
    rnums: Rnums,
    graph: nx.Graph,
    stream: Stream,
):
    """
    Resolves a ordered list of rnum specifications for the specified atom,
    using the specified dictionary of unpaired rnums specifications.
    Bonds are formed in the specified graph as required.

    The dictionary of rnums specifications is changed in-place.

    :param atom_rnums: A dictionary of rnums for the specified atom index,
                       mapping of an integer representing the rnum index
                       to a float, representing the associated bond order
    :type atom_rnum: Rnum
    :param atom_index: The index of the atom associated with :param:`atom_rnum`.
    :type atom_index: int
    :param rnums: A dictionary of unpaired rnum specifications,
                  mapping each rnum to a tuple of the index of the atom
                  where the rnum first occurred, and the specified bond order.
    :param graph: The graph to form bonds on
    :type graph: nx.Graph
    :raises ParserError: If a pair of rnums is found, both explicitly state
                         the bond order, and those orders are mismatched;
                         a pair of rnums attempts to join two already bonded atoms;
                         or a pair of rnums attempts to join an atom to itself.
    """
    for rnum_index, bond_order in atom_rnums:
        if rnum_index in rnums:
            # Rnum index already encountered
            other_atom_index, other_bond_order = rnums[rnum_index]

            # Cannot bond atom to itself
            if atom_index == other_atom_index:
                raise new_exception("Cannot bond atom to itself using rnum", stream)
            # Cannot already bond atoms again
            if graph.has_edge(atom_index, other_atom_index):
                raise new_exception("Atoms already bonded", stream)

            # Aim to store bond order in variable 'bond_order'
            if bond_order is not None and other_bond_order is not None:
                if bond_order != other_bond_order:
                    raise ParserError(
                        f"Explicit bond orders for rnum {rnum_index} do not match: "
                        f"{bond_order} and {other_bond_order}"
                    )
            elif bond_order is None and other_bond_order is not None:
                bond_order = other_bond_order
            elif bond_order is None and other_bond_order is None:
                # Imply bond order from atom types
                if (
                    graph.nodes[atom_index]["aromatic"]
                    and graph.nodes[other_atom_index]["aromatic"]
                ):
                    bond_order = 1.5
                else:
                    bond_order = 1

            graph.add_edge(atom_index, other_atom_index, **new_bond(order=bond_order))
            del rnums[rnum_index]
        else:
            # Rnum index not encountered
            rnums[rnum_index] = (atom_index, bond_order)


def new_atom(**attrs: AtomAttribute) -> Atom:
    """
    Create new atom attributes dictionary from default atom attributes template.

    Keyword arguments can be used to override defaults.
    Raises :class:`KeyError` if any keyword attributes do not exist
    """
    atom = DEFAULT_ATOM.copy()
    for attr, value in attrs.items():
        if attr not in atom:
            raise KeyError(attr)
        atom[attr] = value
    return atom


def new_bond(**attrs: BondAttribute) -> Bond:
    """
    Create new bond attributes dictionary from default bond attributes template.

    Keyword arguments can be used to override defaults.
    Raises :class:`KeyError` if any keyword attributes do not exist
    """
    bond = DEFAULT_BOND.copy()
    for attr, value in attrs.items():
        if attr not in bond:
            raise KeyError(attr)
        bond[attr] = value
    return bond


def catch_stop_iteration(func):
    """
    Decorator for methods that throw :class:`StopIteration`.
    Wraps the method such that the exception is caught, and :class:`ParserError`
    is thrown instead.

    :param func: The function to decorate
    :type func: Callable[[Stream], T]
    :return: The decorated function
    :rtype: Callable[[Stream], T]
    """

    @wraps(func)
    def wrapper(stream, *args, **kwargs) -> T:
        try:
            return func(stream, *args, **kwargs)
        except StopIteration:
            raise new_exception("Unexpected end-of-stream", stream)

    return wrapper


def new_exception(msg: str, stream: Stream, exc_type: E = ParserError) -> E:
    """
    Instantiates a new parser exception (default) or warning with the specified message,
    from the specified stream.

    :param msg: The exception message
    :type msg: str
    :param stream: The stream to use
    :type stream: Stream
    :param exc_type: The exception class to use, defaults to ParserError
    :type exc_type: E, optional
    :return: The new exception
    :rtype: E
    """
    return exc_type(msg, stream.pos)


def expect(
    stream: Stream[str],
    symbols: Iterable[str],
    terminal: Optional[str] = None,
    default: Union[str, object] = EXPECT_DEFAULT,
) -> Union[str, object]:
    """
    Expect the next string in the specified stream to be any character
    from the specified iterable :param:`symbols`, otherwise raise :class:`ParserError`.

    If end-of-stream is reached, then return :param:`default` if specified,
    or raise :class:`ParserError`.

    :param stream: The stream to read from
    :type stream: Stream[str]
    :param symbols: An iterable of symbols to expect
    :type symbols: Iterable[str]
    :param terminal: Name of the terminal to expect, used for error raising
    :type terminal: Optional[str]
    :param default: Value to return if end-of-stream is reached
    :type default: Union[str, object]
    :raises ParserError: If next symbol in stream is not an expected symbol
    :return: The symbol from :param:`symbols` encountered.
    :rtype: Iterable[str]
    """
    try:
        peek = stream.peek()
    except StopIteration:
        if default != EXPECT_DEFAULT:
            return default
        else:
            raise new_exception("Unexpected end-of-stream", stream)

    if peek in symbols:
        return next(stream)

    expected = (
        terminal
        if terminal is not None
        else ", ".join(repr(symbol) for symbol in symbols)
    )
    msg = f"Expected {expected}, got {stream.peek()!r}"
    raise new_exception(msg, stream)


@catch_stop_iteration
def parse_number(stream: Stream[str]) -> int:
    """
    Parse a number (integer) from the specified stream

    :param stream: The stream to read from
    :type stream: Stream[str]
    :raise ParserError: If no number is next in stream
    :return: The parsed number
    :rtype: int
    """
    number = ""
    while True:
        try:
            number += parse_digit(stream)
        except ParserError:
            break

    if not number:
        raise new_exception(f"Expected number, got {stream.peek()}", stream)

    return int(number)


@catch_stop_iteration
def parse_bond(stream: Stream[str]) -> Bond:
    """
    Parses a bond symbol from the specified stream

    :param stream: The stream to read from
    :type stream: Stream[str]
    :raises ParserError: If invalid bond symbol is encountered
    :return: A dictionary of attributes for the parsed bond
    :rtype: Bond
    """
    symbol = expect(stream, BOND_TO_ORDER, "bond")
    return new_bond(order=BOND_TO_ORDER[symbol])


@catch_stop_iteration
def parse_isotope(stream: Stream[str]) -> int:
    """
    Parses an isotope specification from the specified stream

    :param stream: The stream to read from
    :type stream: Stream[str]
    :return: The isotope number parsed
    :rtype: Optional[int]
    """
    return parse_number(stream)


@catch_stop_iteration
def parse_element_symbol(stream) -> str:
    """
    Parses an element symbol from the specified stream

    :param stream: The stream to read from
    :type stream: Stream[str]
    :raises ParserError: If the element symbol is not a known element,
                            or a valid element symbol is not read
    :return: The element parsed
    :rtype: str
    """
    element = expect(stream, ELEMENT_SYMBOL_FIRST_CHARS, "alphabetic character")
    next_char = stream.peek("")
    if next_char and element + next_char in ELEMENT_SYMBOLS:
        element += next(stream)

    if element in ELEMENT_SYMBOLS:
        return element

    raise new_exception(f"Invalid element symbol {element!r}", stream)


@catch_stop_iteration
def parse_digit(stream: Stream[str]) -> str:
    """
    Parses a single digit from the specified stream

    :param stream: The stream to read from
    :type stream: Stream[str]
    :raises ParserError: If character from stream is not a digit
    :return: The digit parsed
    :rtype: str
    """
    return expect(stream, string.digits, "digit")


@catch_stop_iteration
def parse_chiral(stream: Stream[str]) -> Tuple[Optional[str], int]:
    """
    Parses an atomic chirality specification from the specified stream.

    The code is the name given to the two-character alphabetic sequence
    specifying the type of chirality, and the index is the name given
    to the one- or two-digit number that follows the code.

    If shorthand chirality, i.e. @ or @@, is used, the code returned is :data:`None`,
    leaving the code to be determined semantically.

    :param stream: The stream to read from
    :type stream: Stream[str]
    :raises ParserError: If the next symbol in the stream is not '@'
    :return: The chirality specification as a tuple of chirality code and index,
             e.g. ("TH", 1)
    :rtype: Tuple[Optional[str], int]
    """
    expect(stream, "@", "'@'")
    peek = stream.peek("X")

    if peek == "@":
        # @@
        next(stream)
        return None, 2
    if peek not in CHIRALITY_CODES_FIRST_CHARS:
        # @
        return None, 1

    # Code, i.e. TB, AL, SP, TB, OH
    code = next(stream)
    code += stream.peek("")
    if code not in CHIRALITY_CODES:
        raise new_exception(f"Unknown chirality code {code!r}", stream)
    next(stream)

    # Index
    if stream.peek("a") not in string.digits:
        # Must be at least one digit
        raise new_exception("Expected digit", stream)
    index = parse_digit(stream)
    if stream.peek("a") in string.digits:
        # Parse second digit
        index += parse_digit(stream)
    index = int(index)

    if index <= 0 or index > CHIRALITY_RANGES[code]:
        raise new_exception(
            f"Invalid chirality {code}{index}: {index} is out of range", stream
        )

    return code, index


@catch_stop_iteration
def parse_hcount(stream: Stream[str]) -> int:
    """
    Parses hydrogen count from the specified stream

    :param stream: The stream to read from
    :type stream: Stream[str]
    :raises ParserError: If the next symbol in the stream is not 'H'
    :return: The hydrogen count
    :rtype: int
    """
    expect(stream, "H", "'H'")
    if stream.peek("a") in string.digits:
        count = int(parse_digit(stream))
    else:
        count = 1

    return count


@catch_stop_iteration
def parse_charge(stream: Stream[str]) -> int:
    """
    Parses a charge from the specified stream

    :param stream: The stream to read from
    :type stream: Stream[str]
    :return: The charge parsed
    :rtype: int
    """
    sign = expect(stream, CHARGE_SYMBOLS, "charge sign")
    if stream.peek("") == sign:
        next(stream)
        warnings.warn(
            new_exception(
                f"Use of {2 * sign} instead of {sign}2 is deprecated",
                stream,
                ParserWarning,
            )
        )
        return int(sign + "2")
    if stream.peek("a") not in string.digits:
        return int(sign + "1")

    first_digit = parse_digit(stream)
    if stream.peek("a") not in string.digits:
        return int(sign + first_digit)

    second_digit = parse_digit(stream)
    return int(sign + first_digit + second_digit)


@catch_stop_iteration
def parse_class(stream: Stream[str]) -> int:
    """
    Parses an atom class specification from the specified stream.

    :param stream: The stream to read from
    :type stream: Stream[str]
    :raises ParserError: If no atom class specification is found
    :return: The atom class as an :class:`int`
    :rtype: int
    """
    expect(stream, ":", "colon for atom class")
    if not stream.peek() in string.digits:
        raise new_exception("Expected number for atom class", stream) from None
    return parse_number(stream)


@catch_stop_iteration
def parse_bracket_atom(stream: Stream[str]) -> Atom:
    """
    Parses a bracket atom from the specified stream

    :param stream: The stream to read from
    :type stream: Stream[str]
    :raises ParserError: If the opening and closing bracket is not found,
                            or no element is found
    :return: A dictionary of atom attributes
    :rtype: Atom
    """

    attrs = {}

    expect(stream, "[")
    if stream.peek() in string.digits:
        attrs["isotope"] = parse_isotope(stream)
    else:
        attrs["isotope"] = None

    attrs["element"] = parse_element_symbol(stream)
    for attr, parse_method, expected_peeks, default in [
        ("chirality", parse_chiral, "@", None),
        ("hcount", parse_hcount, "H", 0),
        ("charge", parse_charge, CHARGE_SYMBOLS, 0),
        ("class", parse_class, ":", None),
    ]:
        if stream.peek() in expected_peeks:
            attrs[attr] = parse_method(stream)
        else:
            attrs[attr] = default

    expect(stream, "]")

    return attrs


@catch_stop_iteration
def parse_organic_symbol(stream: Stream[str]) -> str:
    """
    Parses an organic subset symbol from the specified stream.

    :param stream: The stream to read from
    :type stream: Stream[str]
    :raises ParserError: If the element symbol is not a known element,
                         not a valid element symbol, or is a valid element symbol
                         that cannot be used in an organic context
    :return: The element parsed
    :rtype: str
    """
    try:
        element = expect(stream, ORGANIC_SYMBOL_FIRST_CHARS, "alphabetic character")
    except ParserError:
        if stream.peek() in ELEMENT_SYMBOLS:
            raise new_exception(
                f"Element symbol {stream.peek()!r} "
                "cannot be used in an organic context",
                stream,
            )
        raise
    next_char = stream.peek("")
    if next_char:
        if element + next_char in ORGANIC_SYMBOLS:
            element += next(stream)
        elif element + next_char in ELEMENT_SYMBOLS:
            raise new_exception(
                f"Element symbol {element + next_char!r} "
                "cannot be used in an organic context",
                stream,
            )

    if element in ORGANIC_SYMBOLS:
        return element
    if element in ELEMENT_SYMBOLS:
        raise new_exception(
            f"Element symbol {element!r} cannot be used in an organic context", stream
        )

    raise new_exception(f"Invalid element symbol {element!r}", stream)


@catch_stop_iteration
def parse_atom(stream: Stream[str]) -> Atom:
    """
    Parses an atom in the specified stream.

    :param stream: The stream to read from
    :type stream: Stream[str]
    :raises ParserError: If no atom is found
    :return: A dictionary of atom attributes
    :rtype: Atom
    """

    # Default atom attributes
    atom = new_atom()

    if stream.peek() == "[":
        # Bracket atom
        attrs = parse_bracket_atom(stream)
        atom.update(**attrs)
    elif stream.peek() in ORGANIC_SYMBOL_FIRST_CHARS:
        # Organic subset symbol
        element = parse_organic_symbol(stream)
        atom["element"] = element
    else:
        raise new_exception("Invalid atom or unknown organic symbol", stream)

    # Deal with aromatic atoms
    if atom["element"].islower():
        atom["element"] = atom["element"].upper()
        atom["aromatic"] = True

    return atom


@catch_stop_iteration
def parse_rnum(stream: Stream[str]) -> int:
    """
    Parses an rnum specification from the specified stream.

    :param stream: The stream to read from
    :type stream: Stream[str]
    :return: The parsed rnum
    :rtype: int
    """
    if stream.peek() in string.digits:
        return int(parse_digit(stream))
    else:
        expect(stream, "%", "percentage sign for rnum")
        return int("".join(parse_digit(stream) for _ in range(2)))


@catch_stop_iteration
def parse_chain(
    stream: Stream[str], prev_is_aromatic: bool = False
) -> Tuple[List[Tuple[Atom, AtomRnums]], List[Bond]]:
    """
    Parses a chain from the specified stream.

    A chain is composed of consecutive bonded atoms,
    (which may be the dot bond) without branching.
    Must be associated with a preceding atom.

    :param stream: The stream to read from
    :type stream: Stream[str]
    :param prev_is_aromatic: Whether the atom preceding this chain
                             is aromatic or not. Defaults to :data:`False`.
    :rtype: bool
    :return: A 2-tuple, with the first element list of tuple of atoms
             and a corresponding dictionary of rnums for that atom.
             The second is a list of bonds, in the same order
             as the atoms.
             The number of atoms is always one more than the number of bonds,
             where rnums about the atom preceding the start of the chain
             is stored in the first element of the atoms list
    :rtype: Tuple[List[Atom], List[Bond]]
    """
    atoms = [({"aromatic": prev_is_aromatic}, [])]
    bonds = []
    while True:
        rnum = None
        dot = False
        if stream.peek("") == ".":
            # Dot bond
            next(stream)
            dot = True
            bond = None
        else:
            # Other bonds
            try:
                bond = parse_bond(stream)
            except ParserError:
                bond = None
        # Rnums invalid after dot bond
        if not dot and stream.peek("a") in string.digits + "%":
            # Rnums
            rnum = parse_rnum(stream)
            # The rnum is associated with the most recently parsed atom
            atoms[-1][1].append((rnum, bond))
        if rnum is None:
            # Add parsed bond (for error detection)
            if bond is not None:
                bonds.append(bond)
            # Atoms
            if not stream.peek("") in {"["} | ORGANIC_SYMBOL_FIRST_CHARS:
                break
            atom = parse_atom(stream)
            if not dot and bond is None:
                bond = infer_bond(atom["aromatic"], atoms[-1][0]["aromatic"])
                bonds.append(bond)
            atoms.append((atom, []))
    if len(bonds) + 1 > len(atoms) or not bonds:
        raise new_exception("Expected atom or rnum", stream)
    return atoms, bonds


@catch_stop_iteration
def parse_branch(
    stream: Stream[str], graph: nx.Graph, prev_atom_idx: int, atom_idx: int, rnums: Rnum
):
    line_encountered = False
    expect(stream, "(", "opening parenthesis for branch")
    while True:
        bond = None
        dot = False
        if stream.peek() in BOND_TO_ORDER:
            # Bond
            bond = parse_bond(stream)
        elif stream.peek() == ".":
            # Dot
            next(stream)
            dot = True

        # Line
        if not stream.peek() in {"["} | ELEMENT_SYMBOL_FIRST_CHARS:
            break
        first_atom_idx, atom_idx = parse_line(stream, graph, atom_idx, rnums)
        line_encountered = True

        # Bond previous atom to first atom in line
        if not dot:
            if bond is None:
                bond = infer_bond(
                    graph.nodes[first_atom_idx]["aromatic"],
                    graph.nodes[prev_atom_idx]["aromatic"],
                )
            graph.add_edge(first_atom_idx, prev_atom_idx, **bond)
        prev_atom_idx = atom_idx - 1
    if not line_encountered:
        raise new_exception("Branch cannot be empty", stream)

    expect(stream, ")", "closing parenthesis for branch")
    return atom_idx


@catch_stop_iteration
def parse_line(
    stream: Stream[str], graph: nx.Graph, atom_idx: int, rnums: Rnums
) -> Tuple[int, int]:
    """
    Parses a line from the specified stream, and extends the specified graph
    with the new line.

    A line is an atom, atoms that follow it and any branches
    that begin at the same level of nesting as the first atom.

    :param stream: The stream to read from
    :type stream: Stream[str]
    :param graph: The graph to add the new nodes to
    :type graph: nx.Graph
    :param atom_idx: The atom index to initially number new nodes from
    :type atom_idx: int
    :param rnums: A dictionary storing unpaired rnums
    :type rnums: Rnums
    :raises ParserError: If no atom for the start of the line is found
    :return: A tuple of the index for the first atom in the line,
             and the new value for the atom index counter
    :rtype: Tuple[int, int]
    """

    graph.add_node(atom_idx, **parse_atom(stream))
    first_atom_idx = last_chain_atom_idx = atom_idx
    atom_idx += 1
    while True:
        peek = stream.peek(None)
        if peek is None:
            break
        if (
            peek in ("[", "%", ".")
            or peek in BOND_TO_ORDER
            or peek in string.digits
            or peek in ORGANIC_SYMBOL_FIRST_CHARS
        ):
            # Chain
            atoms, bonds = parse_chain(stream, graph.nodes[atom_idx - 1]["aromatic"])
            _, first_atom_rnums = atoms.pop(0)

            resolve_rnums(first_atom_rnums, atom_idx - 1, rnums, graph, stream)

            for (atom, atom_rnums), bond in zip(atoms, bonds):
                graph.add_node(atom_idx, **atom)
                if bond is not None:
                    graph.add_edge(last_chain_atom_idx, atom_idx, **bond)

                resolve_rnums(atom_rnums, atom_idx, rnums, graph, stream)
                last_chain_atom_idx = atom_idx
                atom_idx += 1
        elif peek == "(":
            # Branch
            atom_idx = parse_branch(stream, graph, last_chain_atom_idx, atom_idx, rnums)
        else:
            break

    return first_atom_idx, atom_idx


def parse_terminator(stream: Stream[str]):
    """
    Parses a terminator.

    :param stream: The stream to read from
    :type stream: Stream[str]
    :raises ParserError: If terminator is not found, and stream is not at end.
    """
    expect(stream, TERMINATORS, "terminator", None)


def get_remainder(stream: Stream[str]) -> str:
    """
    Exhausts the rest of the specified stream, and returns the string from it.

    :param stream: The stream to read from
    :type stream: Stream[str]
    :return: The string with the remaining characters from the stream
    :rtype: str
    """
    return "".join(stream)


def parse(smiles: str) -> Tuple[nx.Graph, str]:
    """
    Parse the specified SMILES string to produce a graph representation.

    :param smiles: The SMILES string to parse
    :type smiles: str
    :return: A tuple of the graph represented by the SMILES string,
                and the remainder of the SMILEs after the terminator
    :rtype: Tuple[nx.Graph, str]
    """
    stream = Stream(smiles)
    graph = nx.Graph()
    atom_idx = 0
    rnums = {}

    if not smiles:
        raise new_exception("Cannot parse empty string", stream)

    # Syntax parsing + on-the-fly semantics
    parse_line(stream, graph, atom_idx, rnums)
    parse_terminator(stream)

    # Post-parsing semantics

    # Fill implied hydrogens for organic atoms
    fill_hydrogens(graph)
    if rnums:
        # Rnums must be matched
        raise ParserError("Unmatched rnums: " + ", ".join(str(rnum) for rnum in rnums))
    # Determine whether shorthand chirality is tetrahedral or allenal
    imply_shorthand_chirality(graph)

    return graph, get_remainder(stream)



#!/usr/bin/env python3

"""
pydepict.depicter

Depicter for determining the graphical placement of atoms in a molecular graph.

Copyright (c) 2022 William Lee and The University of Sheffield. See LICENSE for details
"""

import random
from collections import defaultdict
from copy import deepcopy
from itertools import combinations, cycle, permutations, product
from typing import Dict, Generator, List, Tuple

import networkx as nx

from .consts import (
    ATOM_PATTERNS,
    CHAIN_PATTERN_UNITS,
    EPSILON,
    SAMPLE_SIZE,
    THIRTY_DEGREES,
    AtomPattern,
    ConstraintsCandidates,
    GraphCoordinates,
    NeighborConstraints,
    NeighborSpec,
)
from .errors import DepicterError
from .models import DepictionConstraints, Matrix, Vector
from .utils import (
    depiction_width,
    is_chain_atom,
    neighbors,
    none_iter,
    prune_hydrogens,
    prune_terminals,
)

__all__ = ["depict"]


def _match_atom_pattern(
    pattern: AtomPattern, neighbor_spec_to_count: Dict[NeighborSpec, int]
) -> bool:
    """
    Returns if an atom pattern matches a neighbor spec count.

    They match if the set of keys are equal, and the number of the vectors
    for each neighbor spec matches the specified count.
    """
    return pattern.keys() == neighbor_spec_to_count.keys() and all(
        len(pattern[key]) == neighbor_spec_to_count[key] for key in pattern
    )


def _find_atom_constraints(
    atom_index: int,
    graph: nx.Graph,
) -> List[Tuple[NeighborConstraints, float]]:
    """
    Retrieves all possible atom constraints
    for the atom with the specified index in the specified graph.
    """
    # Determines element of atom
    element = graph.nodes(data="element")[atom_index]
    # Get neighbor data
    neighbors_idxs = tuple(graph[atom_index])
    if not neighbors_idxs:
        # Atom has no neighbors
        return [({}, 1)]
    neighbor_elements, neighbor_bond_orders = zip(
        *(
            (graph.nodes[u]["element"], graph[u][atom_index]["order"])
            for u in neighbors_idxs
        )
    )
    # TODO: "X" for halogens
    patterns = ATOM_PATTERNS[element if element in ATOM_PATTERNS else None]
    # Determine candidates
    candidates = []
    # Iterate over possibilities of neighbor being connected via any bond
    # or being any element
    for neighbor_elements, neighbor_bond_orders in product(
        none_iter(neighbor_elements), none_iter(neighbor_bond_orders)
    ):
        # Map (element, order) pairs to counts, and to list of indices
        neighbor_spec_to_count = defaultdict(lambda: 0)
        neighbor_spec_to_idxs = defaultdict(lambda: [])
        for neighbor_idx, neighbor_element, neighbor_bond_order in zip(
            neighbors_idxs, neighbor_elements, neighbor_bond_orders
        ):
            neighbor_spec = (neighbor_element, neighbor_bond_order)
            neighbor_spec_to_idxs[neighbor_spec].append(neighbor_idx)
            neighbor_spec_to_count[neighbor_spec] += 1
        for pattern, weight in patterns:
            if not _match_atom_pattern(pattern, neighbor_spec_to_count):
                continue
            new_candidate_patterns = [{}]
            # Iterate over vectors for each neighbor spec
            for neighbor_spec, vectors in pattern.items():
                # Copy current partial new candidates
                prev_candidate_patterns = new_candidate_patterns.copy()
                new_candidate_patterns.clear()
                neighbor_idxs = neighbor_spec_to_idxs[neighbor_spec]
                # Iterate over each partial new candidate
                for prev_candidate_pattern in prev_candidate_patterns:
                    # Iterate over permutations of vectors
                    for neighbor_idxs_perm in permutations(neighbor_idxs):
                        new_candidate_pattern = prev_candidate_pattern.copy()
                        for vector, neighbor_idx in zip(vectors, neighbor_idxs_perm):
                            new_candidate_pattern[neighbor_idx] = vector
                        if new_candidate_pattern not in new_candidate_patterns:
                            new_candidate_patterns.append(new_candidate_pattern)
            # Iterate over candidate patterns
            for new_candidate_pattern in new_candidate_patterns:
                candidate = (new_candidate_pattern, weight)
                if candidate not in candidates:
                    candidates.append(candidate)

    return candidates


def _find_chains(atoms: List[int], graph: nx.Graph) -> List[List[int]]:
    """
    Finds all chains in a graph
    """
    unchained_chain_atoms = [
        atom_index for atom_index in atoms if is_chain_atom(atom_index, graph)
    ]
    chains = []
    while len(unchained_chain_atoms) >= 4:
        possible_chains = []
        for u, v in permutations(unchained_chain_atoms, r=2):
            if not nx.has_path(graph, u, v):
                continue
            possible_chains.extend(nx.all_simple_paths(graph, u, v))
        if not possible_chains:
            # No chains possible
            break
        possible_chains = list(
            filter(
                lambda chain: all(u in unchained_chain_atoms for u in chain),
                possible_chains,
            )
        )
        if not possible_chains:
            # No chains that contains atoms that are not in chains already
            break
        # Select the longest chain
        longest_chain = max(*possible_chains, key=lambda chain: len(chain))
        if len(longest_chain) < 4:
            break
        chains.append(longest_chain)
        for atom_index in longest_chain:
            unchained_chain_atoms.remove(atom_index)

    return chains


def _chain_triplets(
    chain: List[int], graph: nx.Graph
) -> Generator[Tuple[int, int, int, Tuple[int, ...]], None, None]:
    left = None
    root_neighbors = neighbors(chain[0], graph)
    other_chain_atoms = [v for v in root_neighbors if v != chain[1]]
    if other_chain_atoms:
        left = other_chain_atoms[0]
    subs = [v for v in root_neighbors if v not in (left, chain[1])]
    yield left, chain[0], chain[1], subs

    for left, root, right in zip(chain[:-2], chain[1:-1], chain[2:]):
        subs = neighbors(root, graph, (left, right))
        yield left, root, right, subs

    right = None
    root_neighbors = neighbors(chain[-1], graph)
    other_chain_atoms = [v for v in root_neighbors if v != chain[-2]]
    if other_chain_atoms:
        right = other_chain_atoms[0]
    subs = [v for v in root_neighbors if v not in (chain[-2], right)]
    yield chain[-2], chain[-1], right, subs


def _find_chain_constraints(atoms: List[int], graph: nx.Graph) -> ConstraintsCandidates:
    """
    Returns a set of constraints for chains
    """
    chains = _find_chains(atoms, graph)
    candidates: ConstraintsCandidates = {}
    for chain in chains:
        patterns = []
        for pattern_units in CHAIN_PATTERN_UNITS:
            pattern = []
            for (left, _, right, subs), chain_pattern in zip(
                _chain_triplets(chain, graph), cycle(pattern_units)
            ):
                neighbor_constraints = {}
                (
                    neighbor_constraints[left],
                    neighbor_constraints[right],
                ) = chain_pattern[0]
                if subs:
                    sub_vectors = chain_pattern[1][len(subs)]
                    for sub, vector in zip(subs, sub_vectors):
                        neighbor_constraints[sub] = vector
                pattern.append(neighbor_constraints)
            patterns.append(pattern)
        weights = [1 for _ in range(len(CHAIN_PATTERN_UNITS))]
        candidates[tuple(chain)] = (patterns, weights)
        for atom_index in chain:
            atoms.remove(atom_index)

    return candidates


def _remove_conflicting_constraints(
    u: int,
    v: int,
    vector: Vector,
    constraints_candidates: ConstraintsCandidates,
) -> None:
    """
    Removes constraints candidate conflicts.

    Returns whether there are constraints left for all blocks
    """
    for block, (patterns, weights) in constraints_candidates.items():
        if v not in block:
            continue
        index = block.index(v)
        filtered_patterns = []
        filtered_weights = []
        for pattern, weight in zip(patterns, weights):
            neighbor_constraints = pattern[index]
            if u in neighbor_constraints and neighbor_constraints[u] == -vector:
                filtered_patterns.append(pattern)
                filtered_weights.append(weight)
        if not filtered_patterns:
            return False
        constraints_candidates[block] = (filtered_patterns, filtered_weights)

    return True


def _sample_constraints(
    sample: DepictionConstraints,
    constraints_candidates: ConstraintsCandidates,
) -> None:
    """
    Adds non-conflicting constraints to depiction sample
    """
    candidates_copy = deepcopy(constraints_candidates)
    # Shuffle order in which fragments are considered
    blocks = list(constraints_candidates.keys())
    random.shuffle(blocks)
    for block in blocks:
        # Sample constraint for current block
        patterns, weights = candidates_copy[block]
        # Earlier constraint decision does not work with other constraints
        if not patterns or not weights:
            return False
        ((pattern, weight),) = random.choices(list(zip(patterns, weights)), weights)
        # Delete constraints for current block from constraints
        del candidates_copy[block]

        for u, neighbor_constraints in zip(block, pattern):
            sample.weights[u] = weight
            for v, vector in neighbor_constraints.items():
                # Set vector in graph-wide constraints sample
                sample[u, v] = vector
                # Remove conflicting constraints
                constraints_left = _remove_conflicting_constraints(
                    u, v, vector, candidates_copy
                )
                if not constraints_left:
                    return False
    return True


def _apply_depiction_sample(
    graph: nx.Graph, constraints: DepictionConstraints
) -> Dict[int, Vector]:
    """
    Applies constraints to a graph to produce a dictionary mapping atom index
    to the position vector of that atom.
    """
    coordinates = {0: Vector(0, 0)}
    for u, v in nx.dfs_edges(graph, source=0):
        coordinates[v] = coordinates[u] + constraints[u, v]

    return coordinates


def _congestion(sample: Dict[int, Vector], weights: Dict[int, float], graph: nx.Graph):
    """
    Calculates the congestion of the sample.
    """
    congestion = 0
    for component in nx.connected_components(graph):
        for u, v in combinations(component, 2):
            if not graph.has_edge(u, v):
                congestion += 1 / (
                    (Vector.distance(sample[u], sample[v]) + EPSILON) ** 2
                    * (weights[u] if u in weights else 1)
                    * (weights[v] if v in weights else 1)
                )

    return congestion


def _choose_best_sample(
    coordinates_samples_with_weights: List[Tuple[Dict[int, Vector], Dict[int, float]]],
    graph: nx.Graph,
) -> Dict[int, Vector]:
    """
    Selects the best sample from a list of dictionaries of coordinate samples,
    """
    best_sample, _ = min(
        coordinates_samples_with_weights,
        key=lambda sample_with_weight: _congestion(
            sample_with_weight[0], sample_with_weight[1], graph
        ),
    )
    return best_sample


def _maximize_sample_width(sample: Dict[int, Vector]):
    """
    Rotates a depiction sample such that its width is maximized.
    """
    matrices = [Matrix.rotate(THIRTY_DEGREES * i) for i in range(12)]
    widest_sample = max(
        (
            {atom_index: matrix * vector for atom_index, vector in sample.items()}
            for matrix in matrices
        ),
        key=depiction_width,
    )
    sample.update(widest_sample)


def _postprocess_sample(sample: GraphCoordinates):
    """
    Postprocesses a sample dictionary to produce the final depiction
    """
    _maximize_sample_width(sample)


def depict(graph: nx.Graph) -> GraphCoordinates:
    """
    Determines depiction coordinates for the graph, and returns them.

    :param graph: The graph to calculate depiction coordinates for.
    :type graph: nx.Graph
    :return: A dictionary mapping atom index to position vector
    :rtype: GraphCoordinates
    """
    # Makes list of non-hydrogen, non-terminal atoms in the graph
    atoms: List[int] = list(graph.nodes)
    prune_hydrogens(graph, atoms)
    prune_terminals(graph, atoms)

    constraints_candidates: ConstraintsCandidates = {}

    # Determine chain constraints
    chain_constraints = _find_chain_constraints(atoms, graph)
    constraints_candidates.update(chain_constraints)

    # Determine atom constraints
    for atom_index in atoms:
        patterns = _find_atom_constraints(atom_index, graph)
        if not patterns:
            raise DepicterError(
                f"No candidate constraints found for atom with index {atom_index}"
            )
        constraints_candidates[(atom_index,)] = (
            [(neighbor_constraints,) for neighbor_constraints, _ in patterns],
            [weight for _, weight in patterns],
        )

    # Produce constraint samples
    samples: List[DepictionConstraints] = []
    for _ in range(SAMPLE_SIZE):
        sample = DepictionConstraints()
        if _sample_constraints(sample, constraints_candidates):
            samples.append(sample)
    if not samples:
        raise DepicterError("Could not satisfy constraints")
    # Convert constraints to Cartesian coordinates
    coordinates_samples_with_weights = [
        (_apply_depiction_sample(graph, sample), sample.weights) for sample in samples
    ]
    best_sample = _choose_best_sample(coordinates_samples_with_weights, graph)
    # Postprocess constraints
    _postprocess_sample(best_sample)
    return best_sample



#!/usr/bin/env python3

"""
pydepict

A library for parsing and rendering chemical structures from SMILES strings.

Copyright (c) 2022 William Lee and The University of Sheffield. See LICENSE for details
"""


from typing import NamedTuple, Optional

from .depicter import depict
from .parser import parse
from .renderer import Renderer, render

__all__ = [
    "__author__",
    "__copyright__",
    "__license__",
    "__version__",
    "version_info",
    "show",
    "parse",
    "depict",
    "render",
]


class _VersionInfo(NamedTuple):
    major: int
    minor: int
    patch: int
    post: Optional[int] = None

    def __str__(self):
        return f"{self.major}.{self.minor}.{self.patch}" + (
            f".post{self.post}" if self.post is not None else ""
        )


__author__ = "William Lee; The University of Sheffield"
__copyright__ = "Copyright (c) 2022 William Lee and The University of Sheffield"
__license__ = "MIT"

version_info = _VersionInfo(0, 1, 1, post=1)
__version__ = str(version_info)


def show(smiles: str):
    """
    Parses and displays a diagram representing the input SMILES string.

    :param smiles: The SMILES string to parse and display
    :type smiles: str
    """
    graph, _ = parse(smiles)
    positions = depict(graph)
    render(graph, positions, title=smiles)


# Clean up namespace
del NamedTuple, Optional



#!/usr/bin/env python3

"""
pydepict.errors

Custom error classes

Copyright (c) 2022 William Lee and The University of Sheffield. See LICENSE for details
"""


from typing import Optional


class ParserError(Exception):
    """
    Error class for all errors that occur when parsing input
    """

    def __init__(self, msg: str, position: Optional[int] = None) -> None:
        """
        Initialises an instance of a parser exception

        :param msg: The error message
        :type msg: str
        :param position: The position within the stream at which the error occurred,
                         or :data:`None` if not applicable. Defaults to :data:`None`
        :type position: Optional[int]
        """
        super().__init__(
            msg + (f", position {position}" if position is not None else "")
        )
        self.msg = msg
        self.position = position


class ParserWarning(ParserError, Warning):
    """
    Warning class for all warnings that occur when parsing input
    """

    pass


class DepicterError(Exception):
    """
    Class for all depicter-related exceptions
    """

    pass



#!/usr/bin/env python3

"""
pydepict.models

Models for representing data.

Copyright (c) 2022 William Lee and The University of Sheffield. See LICENSE for details
"""

from collections import defaultdict
from math import cos, sin, sqrt
from typing import (
    Dict,
    Generic,
    Iterable,
    List,
    Literal,
    NamedTuple,
    Tuple,
    TypeVar,
    Union,
)

T = TypeVar("T")

__all__ = ["Stream", "Matrix", "Vector"]


class Stream(Generic[T]):
    """
    Stream class for allowing one-item peekahead.

    .. attribute:: pos

        The position within the iterable at which the stream is,
        initially at 0.

        :type: int
    """

    DEFAULT = object()

    def __init__(self, content: Iterable[T]) -> None:
        self._iter = iter(content)
        self._peek = None
        self.pos = 0

    def __iter__(self) -> "Stream":
        return self

    def __next__(self) -> T:
        next_ = self._peek if self._peek is not None else next(self._iter)
        self._peek = None
        self.pos += 1
        return next_

    def peek(self, default: T = DEFAULT) -> T:
        """
        Returns the next item in the stream without advancing the stream.

        If stream is at end then return :param:`default`.

        :param default: Value to return if stream is at end instead
        :type: T
        :return: The next item in the stream
        :rtype: T
        """
        if self._peek is None:
            try:
                self._peek = next(self._iter)
            except StopIteration:
                if default != self.DEFAULT:
                    return default
                raise
        return self._peek


class DepictionConstraints:
    """
    Implements an endpoint order-independent data structure
    for storing chosen constraints, with weights for each atom.

    Endpoints are ordered numerically when setting the constraint vector.
    Vectors are returned in the direction that corresponds with the order
    that the endpoints are presented in.
    """

    def __init__(self):
        self._dict: Dict[int, Dict[int, Vector]] = defaultdict(lambda: {})
        self.weights: Dict[int, float] = {}

    @staticmethod
    def _sort_key(key: Tuple[int, int]) -> Tuple[Tuple[int, int], bool]:
        u, v = key
        if u > v:
            return (v, u), True
        return key, False

    def __contains__(self, key: Tuple[int, int]) -> bool:
        (u, v), _ = self._sort_key(key)
        return u in self._dict and v in self._dict[u]

    def __getitem__(self, key: Tuple[int, int]) -> "Vector":
        (u, v), flipped = self._sort_key(key)
        if self.__contains__((u, v)):
            return -self._dict[u][v] if flipped else self._dict[u][v]
        raise KeyError(key)

    def __setitem__(self, key: Tuple[int, int], value: "Vector"):
        (u, v), flipped = self._sort_key(key)
        self._dict[u][v] = -value if flipped else value

    def __delitem__(self, key: Tuple[int, int]):
        (u, v), _ = self._sort_key(key)
        if self.__contains__((u, v)):
            del self._dict[u][v]
        raise KeyError(key)

    def clear(self):
        """
        Clears all constraints
        """
        self._dict.clear()


class Matrix:
    """
    Represents a 2x2 matrix
    """

    def __init__(self, values: List[List[float]] = None):
        self._list: List[List[float]]
        if values is None:
            self._list = [[0 for _ in range(2)] for _ in range(2)]
        else:
            if len(values) != 2 or any(len(row) != 2 for row in values):
                raise TypeError(
                    "Provided values is not of correct shape. " "Must be 2x2 list"
                )
            self._list = values

    @classmethod
    def rotate(cls, angle: float) -> "Matrix":
        """
        Returns a new :class:`Matrix` representing an anticlockwise rotation
        by :param:`angle` radians.

        Matrix values are rounded to 5 decimal places
        to avoid issues with cos(pi), sin(pi), etc.

        :param angle: The angle of the rotation that the new matrix represents
        :type angle: float
        :return: The matrix representing the rotation
        :rtype: Matrix
        """
        cos_theta = round(cos(angle), 10)
        sin_theta = round(sin(angle), 10)
        return cls(
            [
                [cos_theta, -sin_theta],
                [sin_theta, cos_theta],
            ]
        )

    def __getitem__(self, key: Tuple[int, int]):
        if not isinstance(key, tuple):
            raise TypeError("Key must be tuple")
        if any(v < 0 or v > 2 for v in key):
            raise ValueError("Indices must be between 0 and 2")
        return self._list[key[0]][key[1]]

    def __mul__(self, vector: "Vector") -> "Vector":
        return Vector(*(sum(u * v for u, v in zip(row, vector)) for row in self._list))

    def __str__(self) -> str:
        return str(self._list)

    def __repr__(self) -> str:
        return f"Matrix({self._list})"


class Vector(NamedTuple):
    x: float
    y: float

    @classmethod
    def from_tuple(cls, coords: Tuple[float, float]) -> "Vector":
        return cls(*coords)

    @classmethod
    def _minmax_all(
        cls, vectors: Iterable["Vector"], func: Literal["min", "max"]
    ) -> "Vector":
        if func == "min":
            function = min
        elif func == "max":
            function = max
        else:
            raise ValueError("func must be 'max' or 'min'")
        all_vectors = [vector for vector in vectors]

        folded_x = function((vector.x for vector in all_vectors), default=0)
        folded_y = function((vector.y for vector in all_vectors), default=0)

        return cls(folded_x, folded_y)

    @classmethod
    def max_all(cls, vectors: Iterable["Vector"]) -> "Vector":
        """
        Calculates the vector that has the maximum value of each component
        in the iterable of vectors provided.

        :param vectors: An iterable of vectors
        :type vectors: Iterable[Vector]
        :return: The vector where each component has the maximum for that component
        :rtype: Vector
        """

        return cls._minmax_all(vectors, "max")

    @classmethod
    def min_all(cls, vectors: Iterable["Vector"]) -> "Vector":
        """
        Calculates the vector that has the minimum value of each component
        in the iterable of vectors provided.

        :param vectors: An iterable of vectors
        :type vectors: Iterable[Vector]
        :return: The vector where each component has the minimum for that component
        :rtype: Vector
        """

        return cls._minmax_all(vectors, "min")

    @staticmethod
    def distance(vector1: "Vector", vector2: "Vector") -> float:
        """
        Calculates the distance between two vectors as if they represented positions
        from a fixed origin.
        """
        return sqrt((vector2.x - vector1.x) ** 2 + (vector2.y - vector1.y) ** 2)

    def normal(self) -> "Vector":
        """
        Calculates the normal to this vector.

        :return: The normal
        :rtype: Vector
        """
        return self.__class__(self.y, -self.x)

    def scale_to(self, magnitude: float) -> "Vector":
        """
        Scales this vector to the specified magnitude, and returns the new vector.

        :param magnitude: The magnitude to scale to
        :type magnitude: float
        :return: The scaled vector
        :rtype: Vector
        """
        if isinstance(magnitude, (float, int)):
            curr_magnitude = abs(self)
            scale_factor = magnitude / curr_magnitude
            return self.__class__(self.x * scale_factor, self.y * scale_factor)
        return NotImplemented

    def rotate(self, angle: float) -> "Vector":
        """
        Rotates this vector by :param:`angle` radians anticlockwise

        :param angle: The angle to rotate by
        :type angle: float
        :return: The rotated vector
        :rtype: Vector
        """
        return Matrix.rotate(angle) * self

    def floor(self) -> "Vector":
        """
        Truncates the two components of the vector.

        :return: The new vector with components truncated.
        :rtype: Vector
        """
        return self.__class__(int(self.x), int(self.y))

    def x_reflect(self) -> "Vector":
        """
        Reflects the vector in the x-axis, and returns the reflected vector

        :return: The reflected vector
        :rtype: Vector
        """
        return self.__class__(self.x, -self.y)

    def y_reflect(self) -> "Vector":
        """
        Reflects the vector in the y-axis, and returns the reflected vector

        :return: The reflected vector
        :rtype: Vector
        """
        return self.__class__(-self.x, self.y)

    def copy(self) -> "Vector":
        """
        Returns a copy of this vector.

        :return: A copy of this vector
        :rtype: Vector
        """
        return self.__class__(self.x, self.y)

    def __abs__(self) -> float:
        return sqrt(self.x**2 + self.y**2)

    def __add__(self, other: "Vector") -> "Vector":
        """
        Returns the sum of two vectors
        """
        if isinstance(other, self.__class__):
            return self.__class__(self.x + other.x, self.y + other.y)
        return NotImplemented

    def __sub__(self, other: "Vector") -> "Vector":
        """
        Returns the difference of two vectors
        """
        if isinstance(other, self.__class__):
            return self.__class__(self.x - other.x, self.y - other.y)
        return NotImplemented

    def __mul__(self, other: Union[float, "Vector"]) -> "Vector":
        if isinstance(other, (float, int)):
            return self.__class__(self.x * other, self.y * other)
        elif isinstance(other, self.__class__):
            # Component-wise multiplication
            return self.__class__(self.x * other.x, self.y * other.y)

    def __rmul__(self, other: Union[float, "Vector"]) -> "Vector":
        return self.__mul__(other)

    def __neg__(self) -> "Vector":
        return self.__mul__(-1)

    def __eq__(self, other: "Vector") -> bool:
        if isinstance(other, self.__class__):
            return self.x == other.x and self.y == other.y
        return NotImplemented

    def __str__(self):
        return f"({self.x}, {self.y})"

    def __repr__(self):
        return f"{self.__class__.__name__}({self.x}, {self.y})"



#!/usr/bin/env python3

"""
pydepict.renderer

Renderer for molecular graphs with relative Cartesian coordinates.

Copyright (c) 2022 William Lee and The University of Sheffield. See LICENSE for details
"""

import os
from collections import defaultdict
from copy import deepcopy
from functools import wraps
from math import sqrt
from threading import RLock, Thread
from typing import Any, Callable, Optional

import networkx as nx
import pygame

from .consts import (
    BLACK,
    BOND_WIDTH,
    DISPLAY_BOND_LENGTH,
    FONT_FAMILY,
    FONT_SIZE,
    FRAME_MARGIN,
    SCREENSHOTS_DIR,
    TEXT_MARGIN,
    WHITE,
    WINDOW_TITLE,
    GraphCoordinates,
)
from .models import Vector
from .utils import average_depicted_bond_length, get_datetime_filename

__all__ = ["Renderer", "render"]


class Renderer:
    """
    Renderer class for rendering molecular graphs.

    The renderer takes a molecular graph, and a dictionary of depiction coordinates.
    These coordinates are then scaled and/or translated
    into coordinates on a graphics canvas, then the molecule is rendered.

    .. attribute:: graph

        Instance of a molecular graph to be rendered by the renderer,
        or :data:`None` for no graph.

        :type: Optional[nx.Graph]

    .. attribute:: render_positions

        Dictionary of render positions for each atom in the graph.

        :type: GraphCoordinates

    .. attribute:: redraw

        Whether or not the diagram should be redrawn on the next event loop iteration.
        Set this to :data:`true` whenever any of the depiction coordinates
        in :attr:`graph` are changed, but you do not set :attr:`graph` itself.

        :type: bool
    """

    def __init__(
        self,
        graph: Optional[nx.Graph] = None,
        positions: Optional[GraphCoordinates] = None,
        *,
        title: Optional[str] = None,
    ):
        self._display_lock = RLock()
        self.render_positions: Optional[GraphCoordinates] = None
        self._text_radii = {}
        self.redraw = False  # XXX: Must be before setting graph
        self.title = title
        self._thread = None
        self._event_cbs = defaultdict(lambda: set())
        self.set_structure(graph, positions)

    def __enter__(self) -> "Renderer":
        self.show(False)
        return self

    def __exit__(self, exc_type, exc_val, exc_cb):
        self.close()

    def _with_display_lock(meth):
        """
        Decorator for methods that acquires the display lock
        before calling the wrapped method, and then releases it
        once the wrapped method returns.
        """

        @wraps(meth)
        def wrapper(self: "Renderer", *args, **kwargs):
            with self._display_lock:
                return meth(self, *args, **kwargs)

        return wrapper

    @_with_display_lock
    def set_structure(
        self,
        graph: Optional[nx.Graph] = None,
        depiction_positions: Optional[GraphCoordinates] = None,
    ):
        """
        Sets the graph and corresponding depiction positions dictionary
        for the renderer

        :param graph: The graph that the renderer displays, defaults to :data`None`
        :type graph: Optional[nx.Graph]
        :param depiction_positions: The depiction positions of the graph to render,
                                    defaults to :data:`None`
        :type depiction_positions: Optional[GraphCoordinates]
        """
        self._graph = graph
        self.set_positions(depiction_positions)

    @_with_display_lock
    def set_positions(self, depiction_positions: Optional[GraphCoordinates] = None):
        """
        Sets the depiction positions dictionary of the graph.

        :param depiction_positions: The dictionary of depiction positions,
                                    defaults to :data:`None'
        :type depiction_positions: Optional[GraphCoordinates]
        """
        self._calculate_geometry(depiction_positions)
        self.redraw = True

    @property
    def graph(self) -> nx.Graph:
        """
        The graph that is being rendered. This is a read-only property.
        Use :meth:`set_structure` to set the graph and depiction coordinates.
        """
        return self._graph

    @property
    def title(self) -> str:
        """
        The title to be displayed in the renderer window.

        Setting this property changes the title shown by the window manager.
        """
        return self._title

    @title.setter
    def title(self, title: Optional[str] = None):
        self._title = title
        pygame.display.set_caption(
            (f"{self._title} - " if self._title is not None else "") + WINDOW_TITLE
        )

    def _calculate_geometry(
        self, depiction_positions: Optional[GraphCoordinates] = None
    ):
        # Calculates display coordinates for atoms in the graph,
        # and recalculates the required display size.
        if self._graph is not None and depiction_positions is None:
            raise ValueError(
                "Depiction coordinates cannot be None if graph is not None"
            )

        if self._graph is not None:
            for atom_index in self._graph.nodes:
                if atom_index not in depiction_positions:
                    raise ValueError(f"Position for atom {atom_index} not provided")
            # Calculate scale factor from depiction coordinates to display coordinates
            if self._graph.edges:
                average_bond_length = average_depicted_bond_length(
                    self._graph, depiction_positions
                )
                scale_factor = DISPLAY_BOND_LENGTH / average_bond_length
            else:
                scale_factor = 1

            # Copy depiction positions to render positions
            self.render_positions = deepcopy(depiction_positions)
            # Invert y-coordinate as graphics origin is in top-left hand corner
            for atom_index in self.render_positions:
                self.render_positions[atom_index] *= Vector(1, -1)

            min_all_render_vector = Vector.min_all(self.render_positions.values())
            for atom_index in self._graph.nodes:
                # Normalises depiction coordinates to be non-negative
                self.render_positions[atom_index] -= min_all_render_vector
                # Multiplies by scale factor
                self.render_positions[atom_index] *= scale_factor
                # Adds frame margin
                self.render_positions[atom_index] += Vector(FRAME_MARGIN, FRAME_MARGIN)

            # Calculates display size
            max_all_render_vector = Vector.max_all(self.render_positions.values())
        else:
            self.render_positions = None
            max_all_render_vector = Vector(0, 0)
        self.display_size = tuple(
            max_all_render_vector + Vector(FRAME_MARGIN, FRAME_MARGIN)
        )

    def _display_atom(self, atom_index: int) -> bool:
        """
        Returns whether to render the atom with the given index
        """
        element = self._graph.nodes[atom_index]["element"]
        if element == "C":
            return False
        return True

    @_with_display_lock
    def _render_atom(self, atom_index: int):
        if self._display_atom(atom_index):
            # Skip rendering if atom should not be displayed
            element = self._graph.nodes[atom_index]["element"]
            # Render text from font
            text = self._font.render(element, True, BLACK)
            # Blit text onto canvas, anchored at the center of the text
            x, y = self.render_positions[atom_index]
            coords = (x - text.get_width() / 2, y - self._font.get_ascent() / 2)
            self.display.blit(text, coords)
            # Store radius of rendered text
            self._text_radii[atom_index] = (
                sqrt(text.get_width() ** 2 + self._font.get_height() ** 2) / 2
                + TEXT_MARGIN
            )
        else:
            # Set atom display radius to 0
            self._text_radii[atom_index] = 0

    @_with_display_lock
    def _render_bond(self, u: int, v: int):
        bond_order = self._graph.edges[u, v]["order"]
        # Coordinates for bond endpoints
        coords1, coords2 = self.render_positions[u], self.render_positions[v]
        # Retrieve rendered radius for atoms, including margin
        atom_radius1, atom_radius2 = self._text_radii[u], self._text_radii[v]
        # Get vector between bond endpoints, and its normal
        line_vector = coords2 - coords1
        line_vector_normal = line_vector.normal()
        # Calculate length of bond vector
        atom1_margin_vector = line_vector.scale_to(atom_radius1)
        atom2_margin_vector = line_vector.scale_to(atom_radius2)
        # Calculate ends of bond line
        line_end1 = (coords1 + atom1_margin_vector).floor()
        line_end2 = (coords2 - atom2_margin_vector).floor()

        for i in range(bond_order):
            offset_magnitude = (i / (bond_order - 1) - 1 / 2) if bond_order > 1 else 0
            offset = line_vector_normal.scale_to(offset_magnitude * BOND_WIDTH)
            pygame.draw.aaline(
                self.display, BLACK, line_end1 + offset, line_end2 + offset
            )

    @_with_display_lock
    def _render(self):
        if self.redraw:
            self._text_radii.clear()
            self.display = pygame.display.set_mode(self.display_size)
            # Draw on display
            self.display.fill(WHITE)
            if self._graph is not None:
                for atom_index in self._graph.nodes:
                    self._render_atom(atom_index)
                for u, v in self._graph.edges:
                    self._render_bond(u, v)
            self.redraw = False
        pygame.display.update()

    def _handle_events(self):
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                self._running = False
                break
            elif event.type == pygame.KEYDOWN:
                ctrl = pygame.key.get_mods() & pygame.KMOD_CTRL
                if ctrl and event.key == pygame.K_s:
                    pygame.image.save(
                        self.display,
                        os.path.join(SCREENSHOTS_DIR, get_datetime_filename() + ".png"),
                    )

    def _init(self):
        pygame.init()
        self._running = True
        self._font = pygame.font.SysFont(FONT_FAMILY, size=FONT_SIZE)

    def _loop(self):
        while self._running:
            self._handle_events()
            if not self._running:
                break
            self._render()

        pygame.quit()
        self._fire("close")

    def _fire(self, event: str):
        """
        Fire callbacks for :param:`event`.
        """
        for func in self._event_cbs[event]:
            func()

    def show(self, blocking: bool = True):
        """
        Displays the renderer window.

        This method blocks the calling thread with the event loop,
        unless :param:`blocking` is set to :data:`True`, in which case
        the event loop is called in a separate thread, and the method
        returns after the thread is started.

        :param blocking: Whether or not this method blocks the calling thread.
        :type blocking: bool
        """
        self._init()
        if blocking:
            self._loop()
        else:
            self._thread = Thread(target=self._loop, daemon=True)
            self._thread.start()

    def on(self, event: str, func: Callable[[], Any]):
        """
        Adds a callback function that is called when event
        with name :param:`event` is fired, e.g.::
            renderer.on('close', callback)

        The callback function should have no required arguments,
        and any return value is ignored.

        :param event: The name of the event to bind a callback for
        :type event: str
        :param func: The callback function to add.
        :type func: Callable[[], Any]
        """
        self._event_cbs[event].add(func)

    def not_on(self, event: str, func: Callable[[], Any]):
        """
        Removes a callback function from being called
        when an event with name :param:`event` is fired.

        If the callback function does not exist, then this is ignored.

        :param event: The name of the event to remove
        :type event: str
        :param func: The callback function to remove.
        :type func: Callable[[], Any]
        """
        if func in self._event_cbs[event]:
            self._event_cbs[event].remove(func)

    def close(self):
        """
        Closes the renderer window.
        """
        # pygame quits when the current event loop iteration is completed
        self._running = False
        if self._thread is not None:
            self._thread.join()


def render(
    graph: nx.Graph,
    positions: GraphCoordinates,
    *,
    title: Optional[str] = None,
):
    """
    Shortcut for using :class:`Renderer`. Equivalent to::
        renderer = Renderer(graph, positions, title=title)
        renderer.show()

    This function blocks the current thread until the renderer window is closed.

    :param graph: The graph to render
    :type graph: nx.Graph
    :param positions: The depiction positions of the graph to render
    :type: GraphCoordinates
    :param title: The title to display in the renderer window, defaults to :data:`None`
    :type title: str
    """
    renderer = Renderer(graph, positions, title=title)
    renderer.show()
